@@ -62,14 +62,12 @@ extern "C" {
 #endif
 
 typedef struct _FcitxLibpinyinCandWord {
-    boolean issentence;
     boolean ispunc;
-    phrase_token_t token;
+    int idx;
 } FcitxLibpinyinCandWord;
 
 typedef struct _FcitxLibpinyinFixed {
     int len;
-    phrase_token_t token;
 } FcitxLibpinyinFixed;
 
 CONFIG_DESC_DEFINE(GetLibpinyinConfigDesc, "fcitx-libpinyin.desc")
@@ -80,9 +78,7 @@ static void ConfigLibpinyin(FcitxLibpinyinAddonInstance* libpinyin);
 static int LibpinyinGetOffset(FcitxLibpinyin* libpinyin);
 static void FcitxLibpinyinSave(void *arg);
 static bool LibpinyinCheckZhuyinKey(FcitxKeySym sym, FCITX_ZHUYIN_LAYOUT layout, boolean useTone) ;
-inline char* LibPinyinTransString(FcitxLibpinyin* libpinyin, char* sentence);
 char* LibpinyinGetSentence(FcitxLibpinyin* libpinyin);
-char* LibpinyinTransToken(FcitxLibpinyin* libpinyin, phrase_token_t token);
 
 static const char *input_keys [] = {
      "1qaz2wsxedcrfv5tgbyhnujm8ik,9ol.0p;/-",       /* standard kb */
@@ -117,10 +113,10 @@ bool LibpinyinCheckZhuyinKey(FcitxKeySym sym, FCITX_ZHUYIN_LAYOUT layout, boolea
     const char* tones = tone_keys[layout];
     if (strchr(keys, key))
         return true;
-    
+
     if (useTone && strchr(tones, key))
         return true;
-    
+
     return false;
 }
 
@@ -159,6 +155,12 @@ void FcitxLibpinyinReset (void* arg)
     FcitxLibpinyin* libpinyin = (FcitxLibpinyin*) arg;
     libpinyin->buf[0] = '\0';
     libpinyin->cursor_pos = 0;
+
+    if (libpinyin->candidate) {
+        g_array_free(libpinyin->candidate, TRUE);
+        libpinyin->candidate = NULL;
+    }
+
     if (libpinyin->fixed_string->len > 0)
         g_array_remove_range(libpinyin->fixed_string, 0, libpinyin->fixed_string->len);
     if (libpinyin->inst)
@@ -193,7 +195,7 @@ INPUT_RETURN_VALUE FcitxLibpinyinDoInput(void* arg, FcitxKeySym sym, unsigned in
     FcitxLibpinyin* libpinyin = (FcitxLibpinyin*) arg;
     FcitxLibpinyinConfig* config = &libpinyin->owner->config;
     FcitxInputState* input = FcitxInstanceGetInputState(libpinyin->owner->owner);
-    
+
     if (FcitxHotkeyIsHotKeySimple(sym, state))
     {
         /* there is some special case that ';' is used */
@@ -203,9 +205,9 @@ INPUT_RETURN_VALUE FcitxLibpinyinDoInput(void* arg, FcitxKeySym sym, unsigned in
             || (libpinyin->type == LPT_Zhuyin && LibpinyinCheckZhuyinKey(sym, config->zhuyinLayout, config->useTone))
         )
         {
-            if (strlen(libpinyin->buf) == 0 && sym == ('\'' || sym == ';'))
+            if (strlen(libpinyin->buf) == 0 && (sym == '\'' || sym == ';'))
                 return IRV_TO_PROCESS;
-            
+
             if (strlen(libpinyin->buf) < MAX_PINYIN_INPUT)
             {
                 size_t len = strlen(libpinyin->buf);
@@ -216,9 +218,9 @@ INPUT_RETURN_VALUE FcitxLibpinyinDoInput(void* arg, FcitxKeySym sym, unsigned in
                 libpinyin->buf[len + 1] = 0;
                 libpinyin->buf[libpinyin->cursor_pos] = (char) (sym & 0xff);
                 libpinyin->cursor_pos ++;
-                
+
                 size_t parselen = FcitxLibpinyinParse(libpinyin, libpinyin->buf);
-                
+
                 if (parselen == 0 && strlen(libpinyin->buf) == 1 && libpinyin->type != LPT_Shuangpin
                     && !(libpinyin->type == LPT_Pinyin && !libpinyin->owner->config.incomplete)
                     && !(libpinyin->type == LPT_Zhuyin && !libpinyin->owner->config.chewingIncomplete))
@@ -232,7 +234,7 @@ INPUT_RETURN_VALUE FcitxLibpinyinDoInput(void* arg, FcitxKeySym sym, unsigned in
                 return IRV_DO_NOTHING;
         }
     }
-    
+
     if (FcitxHotkeyIsHotKey(sym, state, FCITX_SPACE)
         || (libpinyin->type == LPT_Zhuyin && FcitxHotkeyIsHotKey(sym, state, FCITX_ENTER)))
     {
@@ -242,17 +244,17 @@ INPUT_RETURN_VALUE FcitxLibpinyinDoInput(void* arg, FcitxKeySym sym, unsigned in
 
         return FcitxCandidateWordChooseByIndex(FcitxInputStateGetCandidateList(input), 0);
     }
-    
+
     if (FcitxHotkeyIsHotKey(sym, state, FCITX_LIBPINYIN_SHIFT_ENTER)) {
         size_t len = strlen(libpinyin->buf);
         if (len == 0)
             return IRV_TO_PROCESS;
-        
+
         strcpy(FcitxInputStateGetOutputString(input), libpinyin->buf);
 
         return IRV_COMMIT_STRING;
     }
-    
+
     if (FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE) || FcitxHotkeyIsHotKey(sym, state, FCITX_DELETE))
     {
         if (strlen(libpinyin->buf) > 0)
@@ -402,30 +404,32 @@ void FcitxLibpinyinLoad(FcitxLibpinyin* libpinyin)
 {
     if (libpinyin->inst != NULL)
         return;
-    
+
     FcitxLibpinyinAddonInstance* libpinyinaddon = libpinyin->owner;
-    
+
     if (libpinyin->type == LPT_Zhuyin && libpinyin->owner->zhuyin_context == NULL) {
         char* user_path = FcitxLibpinyinGetUserPath(libpinyinaddon->config.bSimplifiedDataForZhuyin ? LPLT_Simplified : LPLT_Traditional );
         char* syspath = FcitxLibpinyinGetSysPath(libpinyinaddon->config.bSimplifiedDataForZhuyin ? LPLT_Simplified : LPLT_Traditional );
         libpinyinaddon->zhuyin_context = pinyin_init( syspath, user_path);
+        pinyin_load_phrase_library(libpinyinaddon->zhuyin_context, 15);
         free(user_path);
         free(syspath);
     }
-    
+
     if (libpinyin->type != LPT_Zhuyin && libpinyin->owner->pinyin_context == NULL) {
         char* user_path = FcitxLibpinyinGetUserPath(libpinyinaddon->config.bTraditionalDataForPinyin ? LPLT_Traditional : LPLT_Simplified );
         char* syspath = FcitxLibpinyinGetSysPath(libpinyinaddon->config.bTraditionalDataForPinyin ? LPLT_Traditional : LPLT_Simplified );
         libpinyinaddon->pinyin_context = pinyin_init(syspath, user_path);
+        pinyin_load_phrase_library(libpinyinaddon->pinyin_context, 15);
         free(user_path);
         free(syspath);
     }
-    
+
     if (libpinyin->type == LPT_Zhuyin)
         libpinyin->inst = pinyin_alloc_instance(libpinyinaddon->zhuyin_context);
     else
         libpinyin->inst = pinyin_alloc_instance(libpinyinaddon->pinyin_context);
-    
+
     ConfigLibpinyin(libpinyinaddon);
 }
 
@@ -437,9 +441,9 @@ boolean FcitxLibpinyinInit(void* arg)
         FcitxInstanceSetContext(libpinyin->owner->owner, CONTEXT_ALTERNATIVE_PREVPAGE_KEY, libpinyin->owner->config.hkPrevPage);
         FcitxInstanceSetContext(libpinyin->owner->owner, CONTEXT_ALTERNATIVE_NEXTPAGE_KEY, libpinyin->owner->config.hkNextPage);
     }
-    
+
     FcitxLibpinyinLoad(libpinyin);
-    
+
     return true;
 }
 
@@ -448,7 +452,16 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
     FcitxInstance* instance = libpinyin->owner->owner;
     FcitxInputState* input = FcitxInstanceGetInputState(instance);
     int offset = LibpinyinGetOffset(libpinyin);
-    
+
+    if (libpinyin->type == LPT_Pinyin) {
+        int libpinyinLen = strlen(libpinyin->inst->m_raw_full_pinyin);
+        int fcitxLen = strlen(libpinyin->buf);
+        if (fcitxLen != libpinyinLen) {
+            strcpy(libpinyin->buf, libpinyin->inst->m_raw_full_pinyin);
+            libpinyin->cursor_pos += libpinyinLen - fcitxLen;
+        }
+    }
+
     int pyoffset = LibpinyinGetPinyinOffset(libpinyin);
     if (pyoffset > libpinyin->cursor_pos)
         libpinyin->cursor_pos = pyoffset;
@@ -458,7 +471,7 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
         hzlen = fcitx_utf8_get_nth_char(sentence, offset) - sentence;
     else
         hzlen = strlen(sentence);
-    
+
     if (hzlen > 0) {
         char* buf = (char*) fcitx_utils_malloc0((hzlen + 1) * sizeof(char));
         strncpy(buf, sentence, hzlen);
@@ -466,7 +479,7 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
         FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", buf);
         free(buf);
     }
-    
+
     int charcurpos = hzlen;
 
     int lastpos = pyoffset;
@@ -475,7 +488,7 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
     {
         PinyinKey* pykey = &g_array_index(libpinyin->inst->m_pinyin_keys, PinyinKey, i);
         PinyinKeyPos* pykeypos = &g_array_index(libpinyin->inst->m_pinyin_key_rests, PinyinKeyPos, i);
-        
+
         if (lastpos > 0) {
             FcitxMessagesMessageConcatLast (FcitxInputStateGetPreedit(input), " ");
             if (curoffset < libpinyin->cursor_pos)
@@ -492,7 +505,7 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
             }
         }
         lastpos = pykeypos->m_raw_end;
-        
+
         switch (libpinyin->type) {
             case LPT_Pinyin: {
                 gchar* pystring = pykey->get_pinyin_string();
@@ -521,7 +534,7 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
                         charcurpos += strlen(initial);
                     }
                     FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_CODE, "%s", initial);
-                    
+
                     if (curoffset + 1 <= libpinyin->cursor_pos) {
                         curoffset += 1;
                         charcurpos += strlen(get_middle_string(pykey)) + strlen(get_final_string(pykey));
@@ -542,7 +555,7 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
             case LPT_Zhuyin: {
                 gchar* pystring = pykey->get_chewing_string();
                 FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_CODE, "%s", pystring);
-                
+
                 if (curoffset + pykeypos->length() <= libpinyin->cursor_pos) {
                     curoffset += pykeypos->length();
                     charcurpos += strlen(pystring);
@@ -553,7 +566,7 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
                     size_t len = fcitx_utf8_strlen(pystring);
                     if (pykey->m_tone != CHEWING_ZERO_TONE)
                         len --;
-                    
+
                     if (diff > len)
                         charcurpos += strlen(pystring);
                     else {
@@ -565,14 +578,14 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
             }
         }
     }
-    
+
     int buflen = strlen(libpinyin->buf);
-    
+
     if (lastpos < buflen) {
         FcitxMessagesMessageConcatLast (FcitxInputStateGetPreedit(input), " ");
         if (lastpos < libpinyin->cursor_pos)
             charcurpos ++;
-        
+
         for (int i = lastpos; i < buflen; i ++)
         {
             char temp[2] = {'\0', '\0'};
@@ -587,27 +600,12 @@ void FcitxLibpinyinUpdatePreedit(FcitxLibpinyin* libpinyin, char* sentence)
     FcitxInputStateSetCursorPos(input, charcurpos);
 }
 
-inline char* LibPinyinTransString(FcitxLibpinyin* libpinyin, char* sentence)
-{
-    if (sentence == NULL)
-        return NULL;
-    
-    return sentence;
-}
-
 char* LibpinyinGetSentence(FcitxLibpinyin* libpinyin)
 {
     char* sentence = NULL;
     pinyin_get_sentence(libpinyin->inst, &sentence);
-    
-    return LibPinyinTransString(libpinyin, sentence);
-}
 
-char* LibpinyinTransToken(FcitxLibpinyin* libpinyin, phrase_token_t token)
-{
-    char* sentence = NULL;
-    pinyin_translate_token(libpinyin->inst, token, &sentence);
-    return LibPinyinTransString(libpinyin, sentence);
+    return sentence;
 }
 
 /**
@@ -631,19 +629,14 @@ INPUT_RETURN_VALUE FcitxLibpinyinGetCandWords(void* arg)
     FcitxInputStateSetRawInputBufferSize(input, strlen(libpinyin->buf));
     FcitxInputStateSetShowCursor(input, true);
     FcitxInputStateSetClientCursorPos(input, 0);
-    
+
     if (libpinyin->type == LPT_Zhuyin) {
         FcitxKeyState state = candidateModifierMap[pyConfig->candidateModifiers];
         FcitxCandidateWordSetChooseAndModifier(candList, "1234567890", state);
     }
     else
         FcitxCandidateWordSetChoose(candList, "1234567890");
-    
-    char* sentence = NULL;
-    
-    pinyin_guess_sentence(libpinyin->inst);
-    sentence = LibpinyinGetSentence(libpinyin);
-    
+
     /* add punc */
     if (libpinyin->type == LPT_Zhuyin
         && strlen(libpinyin->buf) == 1
@@ -660,9 +653,7 @@ INPUT_RETURN_VALUE FcitxLibpinyinGetCandWords(void* arg)
         if (result) {
             FcitxCandidateWord candWord;
             FcitxLibpinyinCandWord* pyCand = (FcitxLibpinyinCandWord*) fcitx_utils_malloc0(sizeof(FcitxLibpinyinCandWord));
-            pyCand->issentence = false;
             pyCand->ispunc = true;
-            pyCand->token = 0;
             candWord.callback = FcitxLibpinyinGetCandWord;
             candWord.extraType = MSG_OTHER;
             candWord.owner = libpinyin;
@@ -670,70 +661,49 @@ INPUT_RETURN_VALUE FcitxLibpinyinGetCandWords(void* arg)
             candWord.strExtra = NULL;
             candWord.strWord = strdup(result);
             candWord.wordType = MSG_OTHER;
-        
+
             FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
         }
     }
+    char* sentence = NULL;
 
+    pinyin_guess_sentence(libpinyin->inst);
+    sentence = LibpinyinGetSentence(libpinyin);
     if (sentence) {
-        FcitxCandidateWord candWord;
-        FcitxLibpinyinCandWord* pyCand = (FcitxLibpinyinCandWord*) fcitx_utils_malloc0(sizeof(FcitxLibpinyinCandWord));
-        pyCand->issentence = true;
-        pyCand->ispunc = false;
-        pyCand->token = 0;
-        candWord.callback = FcitxLibpinyinGetCandWord;
-        candWord.extraType = MSG_OTHER;
-        candWord.owner = libpinyin;
-        candWord.priv = pyCand;
-        candWord.strExtra = NULL;
-        candWord.strWord = strdup(sentence);
-        candWord.wordType = MSG_OTHER;
-        
-        FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
-        
         FcitxLibpinyinUpdatePreedit(libpinyin, sentence);
-        
+
         FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", sentence);
+
+        g_free(sentence);
     }
     else {
         FcitxInputStateSetCursorPos(input, libpinyin->cursor_pos);
         FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", libpinyin->buf);
         FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", libpinyin->buf);
     }
-    
-    TokenVector array = g_array_new(FALSE, FALSE, sizeof(phrase_token_t));
-    pinyin_get_candidates(libpinyin->inst, LibpinyinGetOffset(libpinyin), array);
+
+    if (libpinyin->candidate)
+        g_array_free(libpinyin->candidate, TRUE);
+
+    libpinyin->candidate = g_array_new(FALSE, FALSE, sizeof(lookup_candidate_t));
+    pinyin_get_candidates(libpinyin->inst, LibpinyinGetOffset(libpinyin), libpinyin->candidate);
     int i = 0;
-    for (i = 0 ; i < array->len; i ++) {
-        phrase_token_t token = g_array_index(array, phrase_token_t, i);
-        char* tokenstring = LibpinyinTransToken(libpinyin, token);
-        
-        if (tokenstring) {
-            if (sentence && strcmp(sentence, tokenstring) == 0)
-                continue;
-            
-            FcitxCandidateWord candWord;
-            FcitxLibpinyinCandWord* pyCand = (FcitxLibpinyinCandWord*) fcitx_utils_malloc0(sizeof(FcitxLibpinyinCandWord));
-            pyCand->issentence = false;
-            pyCand->ispunc = false;
-            pyCand->token = token;
-            candWord.callback = FcitxLibpinyinGetCandWord;
-            candWord.extraType = MSG_OTHER;
-            candWord.owner = libpinyin;
-            candWord.priv = pyCand;
-            candWord.strExtra = NULL;
-            candWord.strWord = strdup(tokenstring);
-            candWord.wordType = MSG_OTHER;
-            
-            FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
-        }
-        
-        g_free(tokenstring);
+    for (i = 0 ; i < libpinyin->candidate->len; i ++) {
+        lookup_candidate_t token = g_array_index(libpinyin->candidate, lookup_candidate_t, i);
+        FcitxCandidateWord candWord;
+        FcitxLibpinyinCandWord* pyCand = (FcitxLibpinyinCandWord*) fcitx_utils_malloc0(sizeof(FcitxLibpinyinCandWord));
+        pyCand->ispunc = false;
+        pyCand->idx = i;
+        candWord.callback = FcitxLibpinyinGetCandWord;
+        candWord.extraType = MSG_OTHER;
+        candWord.owner = libpinyin;
+        candWord.priv = pyCand;
+        candWord.strExtra = NULL;
+        candWord.strWord = strdup(token.m_phrase_string);
+        candWord.wordType = MSG_OTHER;
+
+        FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
     }
-    
-    g_array_free(array, TRUE);
-    
-    g_free(sentence);
 
     return IRV_DISPLAY_CANDWORDS;
 }
@@ -751,33 +721,20 @@ INPUT_RETURN_VALUE FcitxLibpinyinGetCandWord (void* arg, FcitxCandidateWord* can
     FcitxLibpinyinCandWord* pyCand = (FcitxLibpinyinCandWord*) candWord->priv;
     FcitxInstance* instance = libpinyin->owner->owner;
     FcitxInputState* input = FcitxInstanceGetInputState(instance);
-    
-    if (pyCand->issentence) {
-        char* sentence = LibpinyinGetSentence(libpinyin);
-        if (sentence) {
-            strcpy(FcitxInputStateGetOutputString(input), sentence);
-            pinyin_train(libpinyin->inst);
-        }
-        else
-            strcpy(FcitxInputStateGetOutputString(input), "");
-        
-        return IRV_COMMIT_STRING;
-    } else if (pyCand->ispunc) {
+
+    if (pyCand->ispunc) {
         strcpy(FcitxInputStateGetOutputString(input), candWord->strWord);
         return IRV_COMMIT_STRING;
     } else {
-        pinyin_choose_candidate(libpinyin->inst, LibpinyinGetOffset(libpinyin), pyCand->token);
-        char* tokenstring = LibpinyinTransToken(libpinyin, pyCand->token);
-        
-        if (tokenstring) {
-            FcitxLibpinyinFixed f;
-            f.len = fcitx_utf8_strlen(tokenstring);
-            f.token = pyCand->token;
-            g_array_append_val(libpinyin->fixed_string, f);
-        }
-        
-        g_free(tokenstring);
-        
+        if (!libpinyin->candidate || libpinyin->candidate->len <= pyCand->idx)
+            return IRV_TO_PROCESS;
+        lookup_candidate_t* cand = &g_array_index(libpinyin->candidate, lookup_candidate_t, pyCand->idx);
+        pinyin_choose_candidate(libpinyin->inst, LibpinyinGetOffset(libpinyin), cand);
+
+        FcitxLibpinyinFixed f;
+        f.len = fcitx_utf8_strlen(cand->m_phrase_string);
+        g_array_append_val(libpinyin->fixed_string, f);
+
         int offset = LibpinyinGetOffset(libpinyin);
         if (offset >= libpinyin->inst->m_pinyin_keys->len)
         {
@@ -786,18 +743,19 @@ INPUT_RETURN_VALUE FcitxLibpinyinGetCandWord (void* arg, FcitxCandidateWord* can
             sentence = LibpinyinGetSentence(libpinyin);
             if (sentence) {
                 strcpy(FcitxInputStateGetOutputString(input), sentence);
+                g_free(sentence);
                 pinyin_train(libpinyin->inst);
             }
             else
                 strcpy(FcitxInputStateGetOutputString(input), "");
-            
+
             return IRV_COMMIT_STRING;
         }
-        
+
         int pyoffset = LibpinyinGetPinyinOffset(libpinyin);
         if (pyoffset > libpinyin->cursor_pos)
             libpinyin->cursor_pos = pyoffset;
-        
+
         return IRV_DISPLAY_CANDWORDS;
     }
     return IRV_TO_PROCESS;
@@ -820,6 +778,29 @@ void FcitxLibpinyinDelete(FcitxLibpinyin* libpinyin)
     g_array_free(libpinyin->fixed_string, TRUE);
 }
 
+void* LibpinyinSavePinyinWord(void* arg, FcitxModuleFunctionArg args)
+{
+#if 0
+    FcitxLibpinyinAddonInstance* libpinyinaddon = (FcitxLibpinyinAddonInstance*) arg;
+    FcitxIM* im = FcitxInstanceGetCurrentIM(libpinyinaddon->owner);
+    pinyin_context_t* context = NULL;
+    if (strcmp(im->uniqueName, "pinyin-libpinyin") == 0 ||
+        strcmp(im->uniqueName, "shuangpin-libpinyin") == 0)
+    {
+        context = libpinyinaddon->pinyin_context;
+    }
+    if (!context)
+        return NULL;
+
+    FcitxLibpinyin* libpinyin = (FcitxLibpinyin*) im->klass;
+    import_iterator_t* iter = pinyin_begin_add_phrases(context, 15);
+    char* hz = (char*) args.args[0];
+    pinyin_iterator_add_phrase(iter, hz, libpinyin->inst->m_raw_full_pinyin, -1);
+    pinyin_end_add_phrases(iter);
+#endif
+    return NULL;
+}
+
 /**
  * @brief initialize the extra input method
  *
@@ -832,19 +813,20 @@ void* FcitxLibpinyinCreate (FcitxInstance* instance)
     FcitxLibpinyinAddonInstance* libpinyinaddon = (FcitxLibpinyinAddonInstance*) fcitx_utils_malloc0(sizeof(FcitxLibpinyinAddonInstance));
     bindtextdomain("fcitx-libpinyin", LOCALEDIR);
     libpinyinaddon->owner = instance;
+    FcitxAddon* addon = FcitxAddonsGetAddonByName(FcitxInstanceGetAddons(instance), "fcitx-libpinyin");
 
     if (!LoadLibpinyinConfig(&libpinyinaddon->config))
     {
         free(libpinyinaddon);
         return NULL;
     }
-    
+
     libpinyinaddon->pinyin = FcitxLibpinyinNew(libpinyinaddon, LPT_Pinyin);
     libpinyinaddon->shuangpin = FcitxLibpinyinNew(libpinyinaddon, LPT_Shuangpin);
     libpinyinaddon->zhuyin = FcitxLibpinyinNew(libpinyinaddon, LPT_Zhuyin);
-    
+
     ConfigLibpinyin(libpinyinaddon);
-    
+
     FcitxInstanceRegisterIM(instance,
                     libpinyinaddon->pinyin,
                     "pinyin-libpinyin",
@@ -861,7 +843,7 @@ void* FcitxLibpinyinCreate (FcitxInstance* instance)
                     5,
                     libpinyinaddon->config.bTraditionalDataForPinyin ? "zh_TW" : "zh_CN"
                    );
-    
+
     FcitxInstanceRegisterIM(instance,
                     libpinyinaddon->shuangpin,
                     "shuangpin-libpinyin",
@@ -878,7 +860,7 @@ void* FcitxLibpinyinCreate (FcitxInstance* instance)
                     5,
                     libpinyinaddon->config.bTraditionalDataForPinyin ? "zh_TW" : "zh_CN"
                    );
-    
+
     FcitxInstanceRegisterIM(instance,
                     libpinyinaddon->zhuyin,
                     "zhuyin-libpinyin",
@@ -896,6 +878,8 @@ void* FcitxLibpinyinCreate (FcitxInstance* instance)
                     libpinyinaddon->config.bSimplifiedDataForZhuyin ? "zh_CN" : "zh_TW"
                    );
 
+    AddFunction(addon, (void*) LibpinyinSavePinyinWord);
+
     return libpinyinaddon;
 }
 
@@ -953,6 +937,8 @@ void ConfigLibpinyin(FcitxLibpinyinAddonInstance* libpinyinaddon)
     if (libpinyinaddon->pinyin_context)
         pinyin_set_double_pinyin_scheme(libpinyinaddon->pinyin_context, FcitxLibpinyinTransShuangpinScheme(config->spScheme));
     pinyin::pinyin_option_t settings = 0;
+    settings |= DYNAMIC_ADJUST;
+    settings |= USE_DIVIDED_TABLE | USE_RESPLIT_TABLE;
     for (int i = 0; i <= FCITX_CR_LAST; i ++) {
         if (config->cr[i])
             settings |= FcitxLibpinyinTransCorrection((FCITX_CORRECTION) i);
@@ -961,15 +947,15 @@ void ConfigLibpinyin(FcitxLibpinyinAddonInstance* libpinyinaddon)
         if (config->amb[i])
             settings |= FcitxLibpinyinTransAmbiguity((FCITX_AMBIGUITY) i);
     }
-    
+
     if (config->incomplete) {
         settings |= PINYIN_INCOMPLETE;
     }
-    
+
     if (config->chewingIncomplete) {
         settings |= CHEWING_INCOMPLETE;
     }
-    
+
     if (config->useTone) {
         settings |= USE_TONE;
     }
